#!/usr/bin/env python3

import re
import os
import json
import numpy as np
import pandas as pd
import datatable as dt

from datatable import f
from pathlib import Path

from typing import Dict, Type

ext = "json"

str_package = "mrs"
str_param_file = "mrs_param"
str_structures_file = "mrs_structures"
str_planning_file_id = "mrs_planning"
str_navigation_file_id = "navigation"
str_search_file_id = "search"
str_plan_types = ["plan", "re_plan"]
str_exp_types = ["ideal", "ros"]

str_experiment_planning_versions = [e + "_" + p for p in str_plan_types for e in str_exp_types]
str_planning_files = [str_planning_file_id + "_" + p for p in str_plan_types]
str_navigation_files = [str_navigation_file_id + "_" + ep for ep in str_experiment_planning_versions]
str_search_files = [str_search_file_id + "_" + ep for ep in str_experiment_planning_versions]

path_data_dir = str(Path(os.popen("rospack find " + str_package).read().replace('\n', '') + "/data/results").resolve())


def sort_nicely(unsorted_list):
    """ Sort the given iterable in the way that humans expect."""

    def convert(text):
        return int(text) if text.isdigit() else text

    def alphanum_key(key):
        return [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(unsorted_list, key=alphanum_key)


def add(table: dt.frame, dictionary: Dict, item: str, t: Type = str, row_select=None):
    dict_item = dictionary[item]
    dict_item = t(dict_item)
    if row_select is not None:
        table[row_select, item] = dict_item
    else:
        table[item] = dict_item


def process_results(data_subdir: str):
    print("PROCESSING DATA: %s ..." % data_subdir)
    data_full_path = path_data_dir + "/" + data_subdir
    all_out_files = [str(path.resolve()) for path in Path(data_full_path).rglob("*." + ext)]
    unique_dirs = sort_nicely(np.unique([os.path.dirname(path) for path in all_out_files]).tolist())
    tab_main = dt.Frame()
    for d in unique_dirs:
        exp_id = d.replace(path_data_dir + '/', "")
        print(" ... processing %s ..." % exp_id)
        new_main = dt.Frame([{"id": exp_id}] * len(str_experiment_planning_versions))
        for i, type_plan in enumerate(str_plan_types):
            for j, type_exp in enumerate(str_exp_types):
                idx = i * len(str_exp_types) + j
                new_main[idx, "type_plan"] = type_plan
                new_main[idx, "type_exp"] = type_exp
        new_main["map"] = exp_id.split("/")[1]
        new_main["it"] = int(exp_id.split("/")[-1].split("-")[-1])
        path_param_file = d + "/" + str_param_file + "." + ext
        if os.path.isfile(path_param_file):
            with open(path_param_file, 'r') as file:
                data = json.load(file)
                add(new_main, data, "robot_vis_radius", t=float)
                add(new_main, data, "nav_velocity_lin", t=float)
                add(new_main, data, "nav_velocity_ang", t=float)
        path_structures_file = d + "/" + str_structures_file + "." + ext
        if os.path.isfile(path_structures_file):
            with open(path_structures_file, 'r') as file:
                data = json.load(file)
                add(new_main, data["cover_loc"], "n", t=int)
        for i, str_file in enumerate(str_planning_files):
            row_select = dt.f.type_plan == str_plan_types[i]
            path_file = d + "/" + str_file + "." + ext
            if os.path.isfile(path_file):
                with open(path_file, 'r') as file:
                    data = json.load(file)
                    add(new_main, data, "mlp_problem_type", row_select=row_select)
                    add(new_main, data, "time_cover_loc", row_select=row_select, t=float)
                    add(new_main, data, "time_prep", row_select=row_select, t=float)
                    add(new_main, data, "time_plan", row_select=row_select, t=float)
                    add(new_main, data, "time_re_plan", row_select=row_select, t=float)
        for i, type_plan in enumerate(str_plan_types):
            for j, type_exp in enumerate(str_exp_types):
                row_select = (dt.f.type_plan == str_plan_types[i]) & (dt.f.type_exp == str_exp_types[j])
                idx = i * len(str_exp_types) + j
                path_file = d + "/" + str_navigation_files[idx] + "." + ext
                if os.path.isfile(path_file):
                    with open(path_file, 'r') as file:
                        data = json.load(file)
                        add(new_main, data, "nav_time_total", t=float, row_select=row_select)
                path_file = d + "/" + str_search_files[idx] + "." + ext
                if os.path.isfile(path_file):
                    with open(path_file, 'r') as file:
                        data = json.load(file)
                        add(new_main, data, "search_expected_time", t=float, row_select=row_select)
                        add(new_main, data, "search_areas_rel_lst", t=float, row_select=row_select)
        for i in range(new_main.nrows):
            if new_main[i, :][dt.rowany(dt.math.isna(dt.f[:])), :].nrows == 0:
                tab_main = dt.rbind([tab_main, new_main[i, :]])
    tab_main.to_csv("csv_times/tab_results_all_" + data_subdir + ".csv")


def main():
    process_results("double_sampling")
    process_results("double_sampling_filtered")
    process_results("triangulation_filtered")
    process_results("kazazakis_filtered")
    process_results("proposed_filtered")
    process_results("proposed_triangulation_boosted")


if __name__ == '__main__':
    main()
