#!/usr/bin/env python3

import re
import os
import json
import numpy as np
import pandas as pd
import datatable as dt
import matplotlib.pyplot as plt
import copy as cp
import matplotlib.font_manager

from datatable import f
from pathlib import Path

from typing import Dict, Type


def load_results(path: str, cover_loc_id: str = None):
    tab = dt.fread(path)
    tab["cover_loc"] = cover_loc_id
    return tab


def select_rows(tab, criterion_columns, criterion_values):
    tab_out = dt.Frame()
    if isinstance(criterion_columns, tuple) and len(criterion_columns) == 2:
        for val in criterion_values:
            tab_out = dt.rbind(tab_out, tab[(criterion_columns[0] == val[0]) & (criterion_columns[1] == val[1]), :])
    else:
        for val in criterion_values:
            tab_out = dt.rbind(tab_out, tab[criterion_columns == val, :])
    return tab_out


def get_selecitons(
        exp_types=["ideal", "ros"],
        velocities=[(0.171, 0.312)],  # (0.285, 0.339)
        radii=[3.0, 5.0, 100.0],
        maps=["emp", "pot", "war", "jar"],
        planning_types=["", "-RP"],
        mlp_problems=["TDP", "ATDP", "GSP", "AGSP", "GSP2", "AGSP2"],
        cover_loc=["DS", "DSF", "KAF", "DTF", "WRF", "WRF-DT-F"],
        iters=list(range(1, (20 + 1)))
):
    return exp_types, velocities, radii, maps, planning_types, mlp_problems, cover_loc, iters


def perform_selections(tab, selections):
    exp_types, velocities, radii, maps, planning_types, mlp_problems, cover_loc, iters = selections
    tab = select_rows(tab, f.type_exp, exp_types)
    tab = select_rows(tab, (f.nav_velocity_lin, f.nav_velocity_ang), velocities)
    tab = select_rows(tab, f.robot_vis_radius, radii)
    tab = select_rows(tab, f.map, maps)
    tab = select_rows(tab, f.type_plan, planning_types)
    tab = select_rows(tab, f.mlp_problem_type, mlp_problems)
    tab = select_rows(tab, f.cover_loc, cover_loc)
    tab = select_rows(tab, f.it, iters)
    return tab


def detail_table(label, tab_main, tab_avgs, label_text=None):
    tab_res = tab_avgs[:, ["method_full", "cover_loc", "problem_type"]]
    tab_res["label"] = label
    instances_reorder = [1, 2, 0, 7, 9, 10, 8, 3, 5, 6, 4]
    for inst in [dt.unique(tab_main["instance"]).to_list()[0][i] for i in instances_reorder]:
        tab_res[inst] = 0.0
        for method in dt.unique(tab_res["method_full"]).to_list()[0]:
            val = tab_main[(f.instance == inst) & (f.method_full == method), label].to_list()[0][0]
            tab_res[f.method_full == method, inst] = val
    tab_res["avg"] = 0.0
    for method in dt.unique(tab_res["method_full"]).to_list()[0]:
        tab_res[f.method_full == method, "avg"] = tab_avgs[f.method_full == method, label + "_avg"].to_list()[0][0]
    if label_text:
        tab_res[:, "label"] = label_text
    return tab_res


def save_detail_table(tab, cover_loc, problem_type):
    with open("latex/detail_" + cover_loc + "+" + problem_type + ".tex", "w") as file:
        tab[(f.cover_loc == cover_loc) & (f.problem_type == problem_type), :].to_pandas().to_latex(
            buf=file, index=False, float_format="%.1f", escape=False)


def main():
    # Load all results.
    tab_main = dt.Frame()
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_proposed_filtered.csv", "WRF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_double_sampling.csv", "DS"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_double_sampling_filtered.csv", "DSF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_kazazakis_filtered.csv", "KAF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_triangulation_filtered.csv", "DTF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv_times/tab_results_all_proposed_triangulation_boosted.csv", "WRF-DT-F"))

    # Cosmetics.
    tab_main[f.mlp_problem_type == "tdp", "mlp_problem_type"] = "TDP"
    tab_main[f.mlp_problem_type == "atdp", "mlp_problem_type"] = "ATDP"
    tab_main[f.mlp_problem_type == "gsp", "mlp_problem_type"] = "GSP"
    tab_main[f.mlp_problem_type == "agsp", "mlp_problem_type"] = "AGSP"
    tab_main[f.mlp_problem_type == "gspw2", "mlp_problem_type"] = "GSP2"
    tab_main[f.mlp_problem_type == "agspw2", "mlp_problem_type"] = "AGSP2"
    tab_main[f.map == "empty", "map"] = "emp"
    tab_main[f.map == "potholes", "map"] = "pot"
    tab_main[f.map == "warehouse", "map"] = "war"
    tab_main[f.map == "jari-huge", "map"] = "jar"
    tab_main[f.type_plan == "plan", "type_plan"] = ""
    tab_main[f.type_plan == "re_plan", "type_plan"] = "-RP"
    tab_main = dt.cbind(tab_main[:, {"problem_type": f.mlp_problem_type + f.type_plan}], tab_main)
    tab_main = dt.cbind(tab_main[:, {"method_full": f.cover_loc + "+" + f.problem_type}], tab_main)
    tab_main = dt.cbind(tab_main[:, {"instance": dt.str64(dt.int64(f.robot_vis_radius)) + "-" + f.map}], tab_main)

    # Get the best known results for each map, visibility radius and velocities.
    tab_min = tab_main[:, {"search_expected_time": dt.min(f.search_expected_time)},
              dt.by("type_exp", "robot_vis_radius", "map", "nav_velocity_lin", "nav_velocity_ang")]
    for i in range(tab_min.nrows):
        et_min = tab_min[i, "search_expected_time"]
        row_select = (f.map == tab_min[i, "map"]) & \
                     (f.robot_vis_radius == tab_min[i, "robot_vis_radius"]) & \
                     (f.nav_velocity_lin == tab_min[i, "nav_velocity_lin"]) & \
                     (f.nav_velocity_ang == tab_min[i, "nav_velocity_ang"]) & \
                     (f.type_exp == tab_min[i, "type_exp"])
        tab_main[row_select, "et_gap"] = 100 * (f.search_expected_time - et_min) / et_min

    # Select only results of interest and group by.
    tab_main = perform_selections(tab_main, get_selecitons(exp_types=["ideal"]))
    tab_main = tab_main[:, {
                               "etg_m": dt.mean(f.et_gap),
                               "et_m": dt.mean(f.search_expected_time),
                               "n_m": dt.mean(f.n),
                               "t_nav_m": dt.mean(f.nav_time_total),
                               "t_cov_m": dt.mean(f.time_cover_loc),
                               "t_pre_m": dt.mean(f.time_prep),
                               "t_pla_m": dt.mean(f.time_plan),
                               "t_cov_pre_pla_m": dt.mean(f.time_cover_loc + f.time_prep + f.time_plan),
                               "t_rep_m": dt.mean(f.time_re_plan),
                               "t_total_m": dt.mean(f.time_cover_loc + f.time_prep + f.time_plan + f.time_re_plan)
                           }, dt.by("instance", "method_full", "cover_loc", "problem_type")]
    tab_avgs = tab_main[:, {
                               "etg_m_avg": dt.mean(f.etg_m),
                               "et_m_avg": dt.mean(f.et_m),
                               "n_m_avg": dt.mean(f.n_m),
                               "t_nav_m_avg": dt.mean(f.t_nav_m),
                               "t_cov_m_avg": dt.mean(f.t_cov_m),
                               "t_pre_m_avg": dt.mean(f.t_pre_m),
                               "t_pla_m_avg": dt.mean(f.t_pla_m),
                               "t_cov_pre_pla_m_avg": dt.mean(f.t_cov_pre_pla_m),
                               "t_rep_m_avg": dt.mean(f.t_rep_m),
                               "t_total_m_avg": dt.mean(f.t_total_m)
                           }, dt.by("method_full", "cover_loc", "problem_type")]
    tab_avgs = tab_avgs[:, :, dt.sort("etg_m_avg")]

    tab_res = dt.Frame()
    tab_res = dt.rbind(tab_res, detail_table("t_cov_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{dis}}$"))
    tab_res = dt.rbind(tab_res, detail_table("n_m", tab_main, tab_avgs, label_text="$n=|\Lset|$"))
    tab_res = dt.rbind(tab_res, detail_table("t_pre_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{pre}}$"))
    tab_res = dt.rbind(tab_res, detail_table("t_pla_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{pla}}$"))
    tab_res = dt.rbind(tab_res, detail_table("t_cov_pre_pla_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{dis+pre+pla}}$"))
    tab_res = dt.rbind(tab_res, detail_table("t_rep_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{rep}}$"))
    tab_res = dt.rbind(tab_res, detail_table("t_total_m", tab_main, tab_avgs, label_text="${\\textit{cpu}}_{\\textit{dis+pre+pla+rep}}$"))
    tab_res = dt.rbind(tab_res, detail_table("t_nav_m", tab_main, tab_avgs, label_text="$t_{\\textit{end}}$"))
    tab_res = dt.rbind(tab_res, detail_table("et_m", tab_main, tab_avgs, label_text="$\\texp$"))
    tab_res = dt.rbind(tab_res, detail_table("etg_m", tab_main, tab_avgs, label_text="\%mG"))
    tab_res = tab_res[:, :, dt.sort("method_full")]
    del tab_res["method_full"]
    save_detail_table(tab_res, "WRF-DT-F", "AGSP-RP")
    save_detail_table(tab_res, "WRF-DT-F", "GSP-RP")
    save_detail_table(tab_res, "WRF-DT-F", "AGSP2-RP")
    save_detail_table(tab_res, "WRF-DT-F", "GSP2-RP")
    save_detail_table(tab_res, "WRF-DT-F", "AGSP2-RP")
    save_detail_table(tab_res, "WRF-DT-F", "GSP2-RP")
    save_detail_table(tab_res, "DTF", "AGSP-RP")
    save_detail_table(tab_res, "WRF", "AGSP-RP")
    save_detail_table(tab_res, "DSF", "AGSP-RP")



if __name__ == '__main__':
    main()
