#!/usr/bin/env python3

import re
import os
import json
import numpy as np
import pandas as pd
import datatable as dt
import matplotlib.pyplot as plt
import copy as cp
import matplotlib.font_manager

from datatable import f
from pathlib import Path

from typing import Dict, Type


def load_results(path: str, cover_loc_id: str = None):
    tab = dt.fread(path)
    tab["cover_loc"] = cover_loc_id
    return tab


def select_rows(tab, criterion_columns, criterion_values):
    tab_out = dt.Frame()
    if isinstance(criterion_columns, tuple) and len(criterion_columns) == 2:
        for val in criterion_values:
            tab_out = dt.rbind(tab_out, tab[(criterion_columns[0] == val[0]) & (criterion_columns[1] == val[1]), :])
    else:
        for val in criterion_values:
            tab_out = dt.rbind(tab_out, tab[criterion_columns == val, :])
    return tab_out


def get_selecitons(
        exp_types=["ideal", "ros"],
        velocities=[(0.171, 0.312)],  # (0.285, 0.339)
        radii=[3.0, 5.0, 100.0],
        maps=["emp", "pot", "war", "jar"],
        planning_types=["", "-RP"],
        mlp_problems=["TDP", "ATDP", "GSP", "AGSP", "GSP2", "AGSP2"],
        cover_loc=["DS", "DSF", "KAF", "DTF", "WRF", "WRF-DT-F"],
        iters=list(range(1, (20 + 1)))
):
    return exp_types, velocities, radii, maps, planning_types, mlp_problems, cover_loc, iters


def perform_selections(tab, selections):
    exp_types, velocities, radii, maps, planning_types, mlp_problems, cover_loc, iters = selections
    tab = select_rows(tab, f.type_exp, exp_types)
    tab = select_rows(tab, (f.nav_velocity_lin, f.nav_velocity_ang), velocities)
    tab = select_rows(tab, f.robot_vis_radius, radii)
    tab = select_rows(tab, f.map, maps)
    tab = select_rows(tab, f.type_plan, planning_types)
    tab = select_rows(tab, f.mlp_problem_type, mlp_problems)
    tab = select_rows(tab, f.cover_loc, cover_loc)
    tab = select_rows(tab, f.it, iters)
    return tab


def main():
    # Load all results.
    tab_main = dt.Frame()
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_proposed_filtered.csv", "WRF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_double_sampling.csv", "DS"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_double_sampling_filtered.csv", "DSF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_kazazakis_filtered.csv", "KAF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_triangulation_filtered.csv", "DTF"))
    tab_main = dt.rbind(tab_main, load_results(
        "csv/tab_results_all_proposed_triangulation_boosted.csv", "WRF-DT-F"))

    # Cosmetics.
    tab_main[f.mlp_problem_type == "tdp", "mlp_problem_type"] = "TDP"
    tab_main[f.mlp_problem_type == "atdp", "mlp_problem_type"] = "ATDP"
    tab_main[f.mlp_problem_type == "gsp", "mlp_problem_type"] = "GSP"
    tab_main[f.mlp_problem_type == "agsp", "mlp_problem_type"] = "AGSP"
    tab_main[f.mlp_problem_type == "gspw2", "mlp_problem_type"] = "GSP2"
    tab_main[f.mlp_problem_type == "agspw2", "mlp_problem_type"] = "AGSP2"
    tab_main[f.map == "empty", "map"] = "emp"
    tab_main[f.map == "potholes", "map"] = "pot"
    tab_main[f.map == "warehouse", "map"] = "war"
    tab_main[f.map == "jari-huge", "map"] = "jar"
    tab_main[f.type_plan == "plan", "type_plan"] = ""
    tab_main[f.type_plan == "re_plan", "type_plan"] = "-RP"
    tab_main = dt.cbind(tab_main[:, {"problem_type": f.mlp_problem_type + f.type_plan}], tab_main)
    tab_main = dt.cbind(tab_main[:, {"method_full": f.cover_loc + "+" + f.problem_type}], tab_main)
    tab_main = dt.cbind(tab_main[:, {"instance": dt.str64(dt.int64(f.robot_vis_radius)) + "-" + f.map}], tab_main)

    temp1 = tab_main[
            (f.method_full == "WRF-DT-F+AGSP-RP") &
            (f.robot_vis_radius == 5) &
            (f.type_exp == "ideal") &
            (f.nav_velocity_lin == 0.171) &
            (f.nav_velocity_ang == 0.312), :]
    temp1_best = temp1[:, {"search_expected_time": dt.min(f.search_expected_time)}, dt.by("instance")]
    for i in range(temp1_best.nrows):
        print(temp1[f.search_expected_time == temp1_best[i, "search_expected_time"], ["it", "map", "search_expected_time", "id"]])
    exit(0)

    # Get the best known results for each map, visibility radius and velocities.
    tab_min = tab_main[:, {"search_expected_time": dt.min(f.search_expected_time)},
              dt.by("type_exp", "robot_vis_radius", "map", "nav_velocity_lin", "nav_velocity_ang")]
    for i in range(tab_min.nrows):
        et_min = tab_min[i, "search_expected_time"]
        row_select = (f.map == tab_min[i, "map"]) & \
                     (f.robot_vis_radius == tab_min[i, "robot_vis_radius"]) & \
                     (f.nav_velocity_lin == tab_min[i, "nav_velocity_lin"]) & \
                     (f.nav_velocity_ang == tab_min[i, "nav_velocity_ang"]) & \
                     (f.type_exp == tab_min[i, "type_exp"])
        tab_main[row_select, "et_gap"] = 100 * (f.search_expected_time - et_min) / et_min

    # Save the best known results.
    res_tab_min = tab_min[(f.nav_velocity_lin == 0.171) & (f.nav_velocity_ang == 0.312), :]
    res_tab_min_temp = dt.Frame()
    for i in range(res_tab_min.nrows):
        row_i = res_tab_min[i, :]
        best_time = res_tab_min[i, "search_expected_time"]
        tab_best_method = tab_main[(f.search_expected_time == best_time) & \
                                   (f.map == row_i["map"]) & \
                                   (f.robot_vis_radius == row_i["robot_vis_radius"]) & \
                                   (f.nav_velocity_lin == row_i["nav_velocity_lin"]) & \
                                   (f.nav_velocity_ang == row_i["nav_velocity_ang"]) & \
                                   (f.type_exp == row_i["type_exp"]), :]
        best_methods = str()
        for j in range(tab_best_method.nrows):
            if not best_methods:
                best_methods = tab_best_method[j, "method_full"]
            else:
                best_methods = best_methods + ", " + tab_best_method[j, "method_full"]
        row_i["methods"] = best_methods
        res_tab_min_temp = dt.rbind(res_tab_min_temp, row_i)
    res_tab_min = res_tab_min_temp
    del res_tab_min["nav_velocity_lin"]
    del res_tab_min["nav_velocity_ang"]
    res_tab_min_ideal = res_tab_min[f.type_exp == "ideal", :]
    res_tab_min_temp = dt.Frame()
    for i in range(res_tab_min_ideal.nrows):
        row_i = res_tab_min_ideal[i, :]
        row_i_ros = res_tab_min[(f.type_exp == "ros") & \
                                (f.map == row_i["map"]) & \
                                (f.robot_vis_radius == row_i["robot_vis_radius"]), :]
        ros_time = -1.0
        ros_methods = "-"
        if row_i_ros.nrows != 0:
            ros_time = row_i_ros[0, "search_expected_time"]
            ros_methods = row_i_ros[0, "methods"]
        row_i["ros_time"] = ros_time
        row_i["ros_methods"] = ros_methods
        res_tab_min_temp = dt.rbind(res_tab_min_temp, row_i)
    res_tab_min = res_tab_min_temp
    del res_tab_min["type_exp"]
    res_tab_min_temp = res_tab_min[f.map == "emp", :]
    res_tab_min_temp = dt.rbind(res_tab_min_temp, res_tab_min[f.map == "pot", :])
    res_tab_min_temp = dt.rbind(res_tab_min_temp, res_tab_min[f.map == "war", :])
    res_tab_min_temp = dt.rbind(res_tab_min_temp, res_tab_min[f.map == "jar", :])
    res_tab_min = res_tab_min_temp
    res_tab_min = res_tab_min[:, :, dt.sort("robot_vis_radius", reverse=True)]
    res_tab_min["robot_vis_radius"] = dt.int64(res_tab_min["robot_vis_radius"])
    res_tab_min.names = (
        "Vis. r.", "Map", "[ideal] ET", "[ideal] Methods", "[ros] ET", "[ros] Methods"
    )
    with open("latex/min_et.tex", "w") as file:
        res_tab_min.to_pandas().to_latex(buf=file, index=False, float_format="%.2f")

    tab_ideal_all = perform_selections(tab_main, get_selecitons(exp_types=["ideal"]))
    tab_ideal_all_inst = tab_ideal_all[:, {
                                              "etg_m": dt.mean(f.et_gap),
                                              "etg_b": dt.min(f.et_gap),
                                              "etg_sd": dt.sd(f.et_gap),
                                              "et_m": dt.mean(f.search_expected_time),
                                              "et_b": dt.min(f.search_expected_time),
                                              "et_sd": dt.sd(f.search_expected_time)
                                          }, dt.by("instance", "method_full", "cover_loc", "problem_type")]
    tab_ideal_all_summary = tab_ideal_all_inst[:, {
                                                      "avg": dt.mean(f.etg_m),
                                                      # "etg_b": dt.min(f.et_gap),
                                                      # "etg_sd": dt.sd(f.et_gap),
                                                  }, dt.by("method_full", "cover_loc", "problem_type")]
    tab_ideal_all_summary = tab_ideal_all_summary[:, :, dt.sort("avg")]
    for inst in dt.unique(tab_ideal_all_inst["instance"]).to_list()[0]:
        tab_ideal_all_summary[inst] = 0.0
        for method in dt.unique(tab_ideal_all_summary["method_full"]).to_list()[0]:
            val = tab_ideal_all_inst[(f.instance == inst) & (f.method_full == method), "etg_m"].to_list()[0][0]
            tab_ideal_all_summary[f.method_full == method, inst] = val

    tab_ideal_all_summary = tab_ideal_all_summary[:, [0, 1, 2, 5, 6, 4, 11, 13, 14, 12, 7, 9, 10, 8, 3]]

    tab_ideal_all_summary["num"] = 0
    count = 1
    for i in range(tab_ideal_all_summary.nrows):
        tab_ideal_all_summary[i, "num"] = count
        count += 1
    tab_ideal_all_summary = tab_ideal_all_summary[:, [15, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]]
    del tab_ideal_all_summary["method_full"]
    with open("latex/all_summary.tex", "w") as file:
        tab_ideal_all_summary.to_pandas().to_latex(buf=file, index=False, float_format="%.1f")

    tab_ideal_wrfdtf_summary = tab_ideal_all_summary[f.cover_loc == "WRF-DT-F", :]
    with open("latex/wrfdtf_summary.tex", "w") as file:
        tab_ideal_wrfdtf_summary.to_pandas().to_latex(buf=file, index=False, float_format="%.1f")

    tab_ideal_agsprp_summary = tab_ideal_all_summary[f.problem_type == "AGSP-RP", :]
    with open("latex/agsprp_summary.tex", "w") as file:
        tab_ideal_agsprp_summary.to_pandas().to_latex(buf=file, index=False, float_format="%.1f")

    tab_ideal_ros = perform_selections(tab_main, get_selecitons(exp_types=["ideal"],
                                                                radii=[5.0],
                                                                planning_types=["-RP"],
                                                                mlp_problems=["GSP", "AGSP", "GSP2", "AGSP2"],
                                                                iters=list(range(1, 5 + 1))
                                                                ))
    tab_ros = perform_selections(tab_main, get_selecitons(exp_types=["ros"],
                                                          radii=[5.0],
                                                          planning_types=["-RP"],
                                                          mlp_problems=["GSP", "AGSP", "GSP2", "AGSP2"],
                                                          iters=list(range(1, 5 + 1))
                                                          ))

    print(tab_ideal_ros.nrows)
    print(tab_ros.nrows)

    tab_ideal_ros_temp = dt.Frame()
    tab_ros_temp = dt.Frame()
    for i in range(tab_ideal_ros.nrows):
        inst = tab_ideal_ros[i, "instance"]
        method = tab_ideal_ros[i, "method_full"]
        it = tab_ideal_ros[i, "it"]
        et_ideal = tab_ideal_ros[i, "search_expected_time"]
        et_ros = tab_ros[(f.instance == inst) & (f.method_full == method) & (f.it == it),
                         "search_expected_time"].to_list()[0][0]
        diff = abs(et_ideal - et_ros)
        if not diff >= 50:
            tab_ideal_ros_temp = dt.rbind(
                tab_ideal_ros_temp, tab_ideal_ros[(f.instance == inst) & (f.method_full == method) & (f.it == it), :])
            tab_ros_temp = dt.rbind(
                tab_ros_temp, tab_ros[(f.instance == inst) & (f.method_full == method) & (f.it == it), :])
    tab_ideal_ros = tab_ideal_ros_temp
    tab_ros = tab_ros_temp

    print(tab_ideal_ros.nrows)
    print(tab_ros.nrows)

    tab_ideal_ros_inst = tab_ideal_ros[:, {
                                              "etg_m": dt.mean(f.et_gap),
                                              "etg_b": dt.min(f.et_gap),
                                              "etg_sd": dt.sd(f.et_gap),
                                              "et_m": dt.mean(f.search_expected_time),
                                              "et_b": dt.min(f.search_expected_time),
                                              "et_sd": dt.sd(f.search_expected_time)
                                          }, dt.by("instance", "method_full", "cover_loc", "problem_type")]
    tab_ideal_ros_summary = tab_ideal_ros_inst[:, {
                                                      "avg": dt.mean(f.etg_m),
                                                      # "etg_b": dt.min(f.et_gap),
                                                      # "etg_sd": dt.sd(f.et_gap),
                                                  }, dt.by("method_full", "cover_loc", "problem_type")]
    tab_ideal_ros_summary = tab_ideal_ros_summary[:, :, dt.sort("avg")]
    for inst in dt.unique(tab_ideal_ros_inst["instance"]).to_list()[0]:
        tab_ideal_ros_summary[inst] = 0.0
        for method in dt.unique(tab_ideal_ros_summary["method_full"]).to_list()[0]:
            val = tab_ideal_ros_inst[(f.instance == inst) & (f.method_full == method), "etg_m"].to_list()[0][0]
            tab_ideal_ros_summary[f.method_full == method, inst] = val
    tab_ideal_ros_summary = tab_ideal_ros_summary[:, [0, 1, 2, 4, 6, 7, 5, 3]]
    tab_ideal_ros_summary["num"] = 0
    count = 1
    for i in range(tab_ideal_ros_summary.nrows):
        tab_ideal_ros_summary[i, "num"] = count
        count += 1
    tab_ideal_ros_summary = tab_ideal_ros_summary[:, [8, 0, 1, 2, 3, 4, 5, 6, 7]]
    # del tab_ideal_ros_summary["method_full"]

    print(tab_ideal_ros_summary)

    tab_ros_inst = tab_ros[:, {
                                  "etg_m": dt.mean(f.et_gap),
                                  "etg_b": dt.min(f.et_gap),
                                  "etg_sd": dt.sd(f.et_gap),
                                  "et_m": dt.mean(f.search_expected_time),
                                  "et_b": dt.min(f.search_expected_time),
                                  "et_sd": dt.sd(f.search_expected_time)
                              }, dt.by("instance", "method_full", "cover_loc", "problem_type")]
    tab_ros_summary = tab_ros_inst[:, {
                                          "avg": dt.mean(f.etg_m),
                                          # "etg_b": dt.min(f.et_gap),
                                          # "etg_sd": dt.sd(f.et_gap),
                                      }, dt.by("method_full", "cover_loc", "problem_type")]
    tab_ros_summary = tab_ros_summary[:, :, dt.sort("avg")]
    for inst in dt.unique(tab_ros_inst["instance"]).to_list()[0]:
        tab_ros_summary[inst] = 0.0
        for method in dt.unique(tab_ros_summary["method_full"]).to_list()[0]:
            val = tab_ros_inst[(f.instance == inst) & (f.method_full == method), "etg_m"].to_list()[0][0]
            tab_ros_summary[f.method_full == method, inst] = val
    tab_ros_summary = tab_ros_summary[:, [0, 1, 2, 4, 6, 7, 5, 3]]
    tab_ros_summary["num"] = 0
    count = 1
    for i in range(tab_ros_summary.nrows):
        tab_ros_summary[i, "num"] = count
        count += 1
    tab_ros_summary = tab_ros_summary[:, [8, 0, 1, 2, 3, 4, 5, 6, 7]]
    # del tab_ros_summary["method_full"]

    tab_ideal_ros_summary["num_ros"] = 0
    tab_ideal_ros_summary["avg_ros"] = 0.0
    for method in dt.unique(tab_ros_summary["method_full"]).to_list()[0]:
        tab_ideal_ros_summary[f.method_full == method, "num_ros"] = \
            tab_ros_summary[f.method_full == method, "num"].to_list()[0][0]
        tab_ideal_ros_summary[f.method_full == method, "avg_ros"] = \
            tab_ros_summary[f.method_full == method, "avg"].to_list()[0][0]
    tab_ideal_ros_summary = tab_ideal_ros_summary[:, [0, 9, 1, 2, 3, 4, 5, 6, 7, 8, 10]]

    print(tab_ideal_ros_summary)

    # fig = plt.figure()
    # ax = fig.add_axes([0.25, 0.25, 0.75, 0.75])
    methods = tab_ideal_ros_summary["method_full"].to_list()[0]
    res_ideal = tab_ideal_ros_summary["avg"].to_list()[0]
    res_ros = tab_ideal_ros_summary["avg_ros"].to_list()[0]
    width = 1 / 3
    X = np.arange(len(methods))
    plt.figure(num=None, figsize=(10, 6), dpi=120, facecolor='w', edgecolor='k')
    plt.rc('axes', axisbelow=True)
    # plt.rc('font', **{'family': 'serif', 'serif': ['Computer Modern Roman']})
    plt.rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica'], 'size': 16})
    plt.rc('text', usetex=True)
    # plt.bar(X + 0.00, res_ideal, width=width, color="black")
    # plt.bar(X + width, res_ros, width=width, color="deepskyblue")
    # plt.plot(X + 0.0, res_ideal, ".", linewidth=5, color="black")
    plt.plot(X, res_ideal, linewidth=3, color="black", zorder=1, marker='o')
    # plt.bar(X, res_ideal, width=0.1, color="black", zorder=2)
    plt.plot(X, res_ros, linewidth=3, color="deepskyblue", zorder=3, marker='o')
    # plt.bar(X, res_ros, width=0.1, color="deepskyblue",zorder=4)
    plt.ylim([0.0, 18.0])
    # plt.plot(X + width / 2, res_ideal, linewidth=2, color="black")
    # plt.plot(X + width / 2, res_ros, linewidth=4, color="black")
    # plt.plot(X + width / 2, res_ros, linewidth=2, color="deepskyblue")
    # plt.plot(X + width / 2, res_ros, ".", linewidth=4, color="black")
    # plt.plot(X + width / 2, res_ideal, ".", linewidth=4, color="black")
    plt.xticks(ticks=X, labels=methods, rotation=45, ha="right", rotation_mode="anchor")
    plt.subplots_adjust(left=0.115, right=0.99, bottom=0.3, top=0.98)
    plt.ylabel("Mean gap [\%] averaged over instances")
    plt.legend(["\\textit{Ideal}", "\\textit{ROS}"])
    plt.grid(True, axis="both")
    plt.savefig("ideal-vs-ros.pdf")

    exit(0)

    # Select only results of interest.
    tab_main = perform_selections(tab_main)

    tab_final_map_radius = tab_main[:, {
                                           "etg_m": dt.mean(f.et_gap),
                                           "etg_b": dt.min(f.et_gap),
                                           "etg_sd": dt.sd(f.et_gap),
                                           "et_m": dt.mean(f.search_expected_time),
                                           "et_b": dt.min(f.search_expected_time),
                                           "et_sd": dt.sd(f.search_expected_time)
                                       }, dt.by("map", "robot_vis_radius", "cover_loc", "problem_type")]
    tab_final_map_radius = tab_final_map_radius[:, :, dt.sort("etg_m")]
    tab_final_map_radius = tab_final_map_radius[:, :, dt.sort("robot_vis_radius", reverse=True)]
    tab_final_map_radius = tab_final_map_radius[:, :, dt.sort("map")]

    tab_final_map_radius = tab_final_map_radius[
                           (f.cover_loc == "DS") |
                           (f.cover_loc == "DSF") |
                           (f.cover_loc == "KAF") |
                           (f.cover_loc == "DTF") |
                           (f.cover_loc == "WRF-DT-F") |
                           (f.cover_loc == "WRF"), :]

    tab_final_map_radius = tab_final_map_radius[
                           (f.problem_type == "AGSP2") | (f.problem_type == "AGSPw2+rp") |
                           (f.problem_type == "GSP2") | (f.problem_type == "GSPw2+rp") |
                           (f.problem_type == "GSP") | (f.problem_type == "GSP+rp") |
                           (f.problem_type == "TDP") | (f.problem_type == "ATDP") |
                           (f.problem_type == "AGSP") | (f.problem_type == "AGSP+rp"), :]

    tab_final = tab_final_map_radius[:, {
                                            "m_etg_m": dt.mean(f.etg_m),
                                            "m_etg_b": dt.mean(f.etg_b),
                                            "sd_etg_sd": dt.sd(f.etg_sd),
                                            "m_et_m": dt.mean(f.et_m),
                                            "m_et_b": dt.mean(f.et_b),
                                            "sd_et_sd": dt.sd(f.et_sd)
                                        }, dt.by("cover_loc", "problem_type")]
    tab_final = tab_final[:, :, dt.sort("m_etg_m")]

    tab_final_map_radius.names = (
        "Map", "Vis. r.", "Cover loc.", "Problem type",
        "ET-mG [%]", "ET-bG [%]", "ET-sdG [%]",
        "ET-m [s]", "ET-b [s]", "ET-sd [s]",
    )
    print(tab_final_map_radius)
    tab_final_map_radius.to_csv("tab_final_map_radius.csv")

    tab_final.names = (
        "Cover loc.",
        "Problem type",
        "mean(ET-mG) [%]",
        "mean(ET-bG) [%]",
        "sd(ET-sdG) [%]",
        "mean(ET-m) [%]",
        "mean(ET-b) [%]",
        "sd(ET-sd) [%]"
    )
    print(tab_final)
    tab_final.to_csv("tab_final.csv")


if __name__ == '__main__':
    main()
